# pyke.krb_compiler.scanner_tables.py. This file automatically created by PLY (version 3.3). Don't edit!
_tabversion   = '3.3'
_lextokens    = {'AS_TOK': 1, 'LP_TOK': 1, 'FOREACH_TOK': 1, 'TAKING_TOK': 1, 'PYTHON_TOK': 1, 'NUMBER_TOK': 1, 'DEINDENT_TOK': 1, 'STEP_TOK': 1, 'EXTENDING_TOK': 1, 'ASSERT_TOK': 1, 'ANONYMOUS_VAR_TOK': 1, 'RP_TOK': 1, 'NOTANY_TOK': 1, 'WITHOUT_TOK': 1, 'INDENT_TOK': 1, 'BC_EXTRAS_TOK': 1, 'CODE_TOK': 1, 'REQUIRE_TOK': 1, 'IN_TOK': 1, 'TRUE_TOK': 1, 'PLAN_EXTRAS_TOK': 1, 'NOT_NL_TOK': 1, 'NONE_TOK': 1, 'WHEN_TOK': 1, 'WITH_TOK': 1, 'FALSE_TOK': 1, 'PATTERN_VAR_TOK': 1, 'CHECK_TOK': 1, 'IDENTIFIER_TOK': 1, 'USE_TOK': 1, 'FC_EXTRAS_TOK': 1, 'STRING_TOK': 1, 'FIRST_TOK': 1, 'NL_TOK': 1, 'FORALL_TOK': 1}
_lexreflags   = 0
_lexliterals  = '*:,!.='
_lexstateinfo = {'code': 'exclusive', 'checknl': 'exclusive', 'INITIAL': 'inclusive', 'indent': 'exclusive'}
_lexstatere   = {'checknl': [('(?P<t_checknl_nl>(\\#.*)?(\\r)?\\n)|(?P<t_checknl_other>[^\\#\\r\\n])', [None, ('t_checknl_nl', 'nl'), None, None, ('t_checknl_other', 'other')])], 'code': [('(?P<t_code_string>\'\'\'([^\\\\]|\\\\.)*?\'\'\'|"""([^\\\\]|\\\\.)*?"""|\'([^\'\\\\\\n\\r]|\\\\.|\\\\(\\r)?\\n)*?\'|"([^"\\\\\\n\\r]|\\\\.|\\\\(\\r)?\\n)*?")|(?P<t_code_comment>[ \\t\\f\\r]*\\#.*)|(?P<t_code_plan>\\$\\$)|(?P<t_code_pattern_var>\\$[a-zA-Z_][a-zA-Z0-9_]*\\b)|(?P<t_code_continuation>\\\\(\\r)?\\n)|(?P<t_code_open>[{([])|(?P<t_code_close>[]})])|(?P<t_code_symbol>[0-9a-zA-Z_]+)|(?P<t_code_space>[ \\t]+)|(?P<t_code_other>[^][(){}$\\\\\'"\\r\\n0-9a-zA-Z_ \\t]+)|(?P<t_code_NL_TOK>(\\r)?\\n([ \\t]*(\\#.*)?(\\r)?\\n)*[ \\t]*)', [None, ('t_code_string', 'string'), None, None, None, None, None, None, ('t_code_comment', 'comment'), ('t_code_plan', 'plan'), ('t_code_pattern_var', 'pattern_var'), ('t_code_continuation', 'continuation'), None, ('t_code_open', 'open'), ('t_code_close', 'close'), ('t_code_symbol', 'symbol'), ('t_code_space', 'space'), ('t_code_other', 'other'), ('t_code_NL_TOK', 'NL_TOK')])], 'INITIAL': [('(?P<t_continuation>\\\\(\\r)?\\n)|(?P<t_NL_TOK>(\\r)?\\n([ \\t]*(\\#.*)?(\\r)?\\n)*)|(?P<t_tsqstring>[uU]?[rR]?\'\'\'([^\\\\]|\\\\.)*?\'\'\')|(?P<t_tdqstring>[uU]?[rR]?"""([^\\\\]|\\\\.)*?""")|(?P<t_sqstring>[uU]?[rR]?\'([^\'\\\\\\n\\r]|\\\\.|\\\\(\\r)?\\n)*?\')|(?P<t_dqstring>[uU]?[rR]?"([^"\\\\\\n\\r]|\\\\.|\\\\(\\r)?\\n)*?")|(?P<t_ANONYMOUS_VAR_TOK>\\$_([a-zA-Z_][a-zA-Z0-9_]*)?)|(?P<t_PATTERN_VAR_TOK>\\$[a-zA-Z][a-zA-Z0-9_]*)|(?P<t_IDENTIFIER_TOK>[a-zA-Z_][a-zA-Z0-9_]*)|(?P<t_float>[-+]?([0-9]+(\\.[0-9]*([eE][-+]?[0-9]+)?|[eE][-+]?[0-9]+)|\\.[0-9]+([eE][-+]?[0-9]+)?))|(?P<t_hexint>[-+]?0[xX][0-9a-fA-F]+)|(?P<t_octalint>[-+]?0[0-7]*)|(?P<t_int>[-+]?[1-9][0-9]*)|(?P<t_LB_TOK>\\[)|(?P<t_LC_TOK>\\{)|(?P<t_LP_TOK>\\()|(?P<t_RB_TOK>\\])|(?P<t_RC_TOK>\\})|(?P<t_RP_TOK>\\))|(?P<t_ignore_comment>\\#.*)', [None, ('t_continuation', 'continuation'), None, ('t_NL_TOK', 'NL_TOK'), None, None, None, None, ('t_tsqstring', 'tsqstring'), None, ('t_tdqstring', 'tdqstring'), None, ('t_sqstring', 'sqstring'), None, None, ('t_dqstring', 'dqstring'), None, None, ('t_ANONYMOUS_VAR_TOK', 'ANONYMOUS_VAR_TOK'), None, ('t_PATTERN_VAR_TOK', 'PATTERN_VAR_TOK'), ('t_IDENTIFIER_TOK', 'IDENTIFIER_TOK'), ('t_float', 'float'), None, None, None, None, ('t_hexint', 'hexint'), ('t_octalint', 'octalint'), ('t_int', 'int'), ('t_LB_TOK', 'LB_TOK'), ('t_LC_TOK', 'LC_TOK'), ('t_LP_TOK', 'LP_TOK'), ('t_RB_TOK', 'RB_TOK'), ('t_RC_TOK', 'RC_TOK'), ('t_RP_TOK', 'RP_TOK'), (None, None)])], 'indent': [('(?P<t_indent_sp>\\n[ \\t]*)', [None, ('t_indent_sp', 'sp')])]}
_lexstateignore = {'code': '', 'checknl': ' \t', 'INITIAL': ' \t', 'indent': ''}
_lexstateerrorf = {'checknl': 't_ANY_error', 'code': 't_ANY_error', 'INITIAL': 't_ANY_error', 'indent': 't_ANY_error'}



# $Id: knowledge_base.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

class gen_tuple(object):
    def __init__(self, tup): self.tup = tup

    def __enter__(self): return self.tup

    def __exit__(self, type, value, tb): pass

Gen_empty = gen_tuple(())
Gen_once = gen_tuple((None,))

class knowledge_base(object):
    ''' This object is a master repository for knowledge entities of different
        names.  These knowledge entities could be facts or rules.  The
        cumulative information maintained in a knowledge_base represents all
        knowledge within a specific domain.
        
        In the syntax: "name1.name2(arg_pattern...)", the knowledge_base name
        is "name1".
    '''
    def __init__(self, engine, name, entity_list_type = None, register = True):
        self.name = name
        self.entity_lists = {}          # {name: entity_list}
        self.entity_list_type = entity_list_type
        self.initialized = False        # used by self.init2
        if register: self.register(engine)
        else: self.engine = engine

    def register(self, engine):
        r'''
            Called at most once either from __init__ or after loading from a
            pickle.
        '''
        self.engine = engine
        name = self.name
        if name in engine.knowledge_bases:
            raise AssertionError("knowledge_base %s already exists" % name)
        if name in engine.rule_bases:
            raise AssertionError("name clash between %s '%s' and "
                                 "rule_base '%s'" %
                                     (self.__class__.__name__, name, name))
        engine.knowledge_bases[name] = self

    def __getstate__(self):
        r'''
            User must call 'register' on the new instance after loading it
            from the pickle.  We do this so that we don't end up pickling the
            whole engine!
        '''
        ans = vars(self).copy()
        del ans['engine']
        return ans

    def init2(self):
        ''' overridden by subclasses. '''
        pass

    def reset(self):
        for entity in self.entity_lists.values(): entity.reset()

    def __repr__(self): return "<%s %s>" % (self.__class__.__name__, self.name)

    def get_entity_list(self, entity_name):
        ans = self.entity_lists.get(entity_name)
        if ans is None:
            if self.entity_list_type:
                ans = self.entity_lists[entity_name] \
                    = self.entity_list_type(entity_name)
            else:
                raise KeyError("%s not found in knowledge_base %s" %
                               (entity_name, self.name))
        return ans

    def lookup(self, bindings, pat_context, entity_name, patterns):
        entity = self.entity_lists.get(entity_name)
        if entity is None: return Gen_empty
        return entity.lookup(bindings, pat_context, patterns)

    def prove(self, bindings, pat_context, entity_name, patterns):
        entity = self.entity_lists.get(entity_name)
        if entity is None: return Gen_empty
        return entity.prove(bindings, pat_context, patterns)

    def add_fc_rule_ref(self, entity_name, fc_rule, foreach_index):
        self.get_entity_list(entity_name) \
            .add_fc_rule_ref(fc_rule, foreach_index)

class knowledge_entity_list(object):
    ''' This object keeps track of all of the knowledge entities sharing the
        same name.  For example, these knowledge entities could be all the
        facts of the same name or all of the rules of the same name.
        Generally, all of the entities in this list may come to bear on
        looking up or proving a single fact or goal.

        In the syntax: "name1.name2(arg_pattern...)", the knowledge entity
        name is "name2".
    '''
    def __init__(self, name):
        self.name = name

    def __repr__(self): return "<%s %s>" % (self.__class__.__name__, self.name)

    def reset(self):
        pass

    def prove(self, bindings, pat_context, patterns):
        return self.lookup(bindings, pat_context, patterns)

    def add_fc_rule_ref(self, fc_rule, foreach_index):
        pass




# $Id: question_base.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import contextlib
from pyke import unique
from pyke import knowledge_base

class question_base(knowledge_base.knowledge_base):
    r'''
        Each instance keeps track of a related set of questions.
    '''
    def __init__(self, name):
        r'''
            This is only used by the compiler, so only creates an instance
            suitable for pickling.

            Specifically, this means that the self.engine is just set to None
            and the instance is not registered with any engine.
        '''
        super(question_base, self).__init__(None, name, register=False)

    def add_question(self, question):
        name = question.name
        if name in self.entity_lists:
            raise AssertionError("question_base %s: duplicate question, %s" %
                                    (self.name, name))
        self.entity_lists[name] = question
        question.set_knowledge_base(self)

    def get_ask_module(self):
        if hasattr(self, 'ask_module'): return self.ask_module
        return self.engine.get_ask_module()

class question(knowledge_base.knowledge_entity_list):
    r'''
        This represents one question in a question_base.  It takes care of
        lookup parameters and caching and delegates the work of actually
        asking the user a question to the user_question object by calling its
        'ask' method passing the format parameters.
    '''
    not_found = unique.unique('question.not_found')

    def __init__(self, name, params, answer_param, user_question):
        super(question, self).__init__(name)
        self.params = tuple(params)
        self.answer_param = answer_param
        try:
            self.answer_param_position = list(params).index(answer_param)
        except ValueError:
            raise ValueError("question %s: answer parameter, %s, "
                             "not in params list: %s" % (answer_param, params))
        self.input_param_positions = \
            tuple([i for i in list(range(len(self.params))) if i != self.answer_param_position])
        self.user_question = user_question
        self.cache = {}

    def __repr__(self):
        return "<question %s(%s): $%s = %s>" % \
               (self.name, ', '.join('$' + p for p in self.params),
                self.answer_param, repr(self.user_question))

    def set_knowledge_base(self, question_base):
        self.knowledge_base = question_base
        self.user_question.set_question_base(question_base)

    def lookup(self, bindings, pat_context, patterns):
        input_params = tuple((self.params[i],
                                str(patterns[i].as_data(pat_context)))
                             for i in self.input_param_positions)
        format_params = dict(input_params)
        ans = self.cache.get(input_params, self.not_found)
        if ans is self.not_found:
            ans = self.cache[input_params] = \
                self.user_question.ask(format_params)

        def gen():
            mark = bindings.mark(True)
            end_done = False
            try:
                if patterns[self.answer_param_position] \
                       .match_data(bindings, pat_context, ans):
                    bindings.end_save_all_undo()
                    end_done = True
                    yield
            finally:
                if not end_done: bindings.end_save_all_undo()
                bindings.undo_to_mark(mark)

        return contextlib.closing(gen())

    def reset(self):
        self.cache.clear()




# $Id: kfbparser.py 49507964ae64 2010-03-27 mtnyogi $
# coding=utf-8
# 
# Copyright © 2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

""" See http://www.dabeaz.com/ply/ply.html for syntax of grammer definitions.
""" 


import os, os.path
from pyke.krb_compiler.ply import yacc
from pyke.krb_compiler import scanner
from pyke import fact_base

tokens = scanner.kfb_tokens

def p_file(p):
    ''' file : nl_opt facts_opt
        facts_opt :
        facts_opt : facts nl_opt
        facts : fact
        facts : facts NL_TOK fact
    '''
    pass

def p_fact0(p):
    ''' fact : IDENTIFIER_TOK LP_TOK RP_TOK '''
    Fact_base.add_universal_fact(p[1], ())

def p_fact1(p):
    ''' fact : IDENTIFIER_TOK LP_TOK data_list RP_TOK '''
    Fact_base.add_universal_fact(p[1], tuple(p[3]))

def p_none(p):
    ''' data : NONE_TOK
        comma_opt :
        comma_opt : ','
        nl_opt :
        nl_opt : NL_TOK
    '''
    p[0] = None

def p_number(p):
    ''' data : NUMBER_TOK
    '''
    p[0] = p[1]

def p_string(p):
    ''' data : STRING_TOK
    '''
    p[0] = eval(p[1])

def p_quoted_last(p):
    ''' data : IDENTIFIER_TOK
    '''
    p[0] = p[1]

def p_false(p):
    ''' data : FALSE_TOK
    '''
    p[0] = False

def p_true(p):
    ''' data : TRUE_TOK
    '''
    p[0] = True

def p_empty_tuple(p):
    ''' data : LP_TOK RP_TOK
    '''
    p[0] = ()

def p_start_list(p):
    ''' data_list : data
    '''
    p[0] = [p[1]]

def p_append_list(p):
    ''' data_list : data_list ',' data
    '''
    p[1].append(p[len(p)-1])
    p[0] = p[1]

def p_tuple(p):
    ''' data : LP_TOK data_list comma_opt RP_TOK '''
    p[0] = tuple(p[2])

def p_error(t):
    if t is None:
        raise SyntaxError("invalid syntax", scanner.syntaxerror_params())
    else:
        raise SyntaxError("invalid syntax",
                          scanner.syntaxerror_params(t.lexpos, t.lineno))

parser = None

def init(this_module, check_tables = False, debug = 0):
    global parser
    if parser is None:
        outputdir = os.path.dirname(this_module.__file__)
        if debug:
            parser = yacc.yacc(module=this_module, write_tables=0,
                               debug=debug, debugfile='kfbparser.yacc.out',
                               outputdir=outputdir)
        else:
            if check_tables:
                kfbparser_mtime = os.path.getmtime(this_module.__file__)
                tables_name = os.path.join(outputdir, 'kfbparser_tables.py')
                try:
                    ok = os.path.getmtime(tables_name) >= kfbparser_mtime
                except OSError:
                    ok = False
                if not ok:
                    #print "regenerating kfbparser_tables"
                    try: os.remove(tables_name)
                    except OSError: pass
                    try: os.remove(tables_name + 'c')
                    except OSError: pass
                    try: os.remove(tables_name + 'o')
                    except OSError: pass
            parser = yacc.yacc(module=this_module, debug=0,
                               optimize=1, write_tables=1,
                               tabmodule='pyke.krb_compiler.kfbparser_tables',
                               outputdir=outputdir)

# Use the first line for normal use, the second for testing changes in the
# grammer (the first line does not report grammer errors!).
def parse(this_module, filename, check_tables = False, debug = 0):
#def parse(this_module, filename, check_tables = False, debug = 1):
    '''
        >>> from pyke.krb_compiler import kfbparser
        >>> kfbparser.parse(kfbparser,
        ...                 os.path.join(os.path.dirname(__file__),
        ...                              'TEST/kfbparse_test.kfb'),
        ...                 True)
        <fact_base kfbparse_test>
    '''
    global Fact_base
    init(this_module, check_tables, debug)
    name = os.path.basename(filename)[:-4]
    Fact_base = fact_base.fact_base(None, name, False)
    with open(filename) as f:
        scanner.init(scanner, debug, check_tables, True)
        scanner.lexer.lineno = 1
        scanner.lexer.filename = filename
        #parser.restart()
        parser.parse(f.read(), lexer=scanner.lexer, tracking=True, debug=debug)
    ans = Fact_base
    Fact_base = None
    return ans




# $Id: ask_wx.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

r'''
    A "match" here is one of:
        - an instance of qa_helpers.regexp
            - msg (for error message)
            - prompt (without the [])
            - match(str) returns converted value (None if no match)
        - an instance of qa_helpers.qmap
            - test (a match)
            - value (value to use)
        - an instance of slice (step must be None)
        - a tuple of matches (implied "or")
        - some other python value (which stands for itself)

    A "review" here is a tuple of (match, review_string)

    "Alternatives" here is a tuple of (tag, label_string)
'''

import sys
if __name__ != "__main__" \
   and ('__main__' not in sys.modules
        or 'doctest' not in sys.modules['__main__'].__file__):
    # So we don't screw up doctest runs on boxes that don't have wxPython
    # installed...
    import wx
import itertools
from pyke import qa_helpers

def review_ans(dlg, ans, review=None):
    if review:

        def matches2(ans, test):
            try:
                qa_helpers.match(ans, test)
                return True
            except ValueError:
                return False

        def matches(ans, test):
            if isinstance(ans, (tuple, list)):
                return any(map(lambda elem: matches2(elem, test),
                                          ans))
            return matches2(ans, test)
        msg = '\n\n'.join(review_str for review_test, review_str in review
                                       if matches(ans, review_test))
        if msg:
            dlg2 = wx.MessageDialog(dlg, msg, 'Answer Information',
                                    wx.OK | wx.ICON_INFORMATION)
            dlg2.ShowModal()
            dlg2.Destroy()

def get_answer(question, title, conv_fn=None, test=None, review=None):
    dlg = wx.TextEntryDialog(None, question, title, '',
                             wx.CENTRE | wx.RESIZE_BORDER | wx.OK)
    while True:
        status = dlg.ShowModal()
        if status != wx.ID_OK:
            raise AssertionError("dlg.ShowModal failed with %d" % status)
        ans = dlg.GetValue()
        try:
            if conv_fn: ans = conv_fn(ans)
            if test: ans = qa_helpers.match(ans, test)
            break
        except ValueError as e:
            err = wx.MessageDialog(dlg,
                                   "Answer should be %s\nGot %s" %
                                       (e.message, repr(ans)),
                                   "Error Notification",
                                   wx.OK | wx.ICON_ERROR)
            err.ShowModal()
            err.Destroy()
            dlg.SetValue("")
    review_ans(dlg, ans, review)
    dlg.Destroy()
    return ans

def ask_yn(question, review=None):
    dlg = wx.MessageDialog(None, question, 'User Question',
                           wx.YES_NO | wx.ICON_QUESTION)

    status = dlg.ShowModal()
    if status not in (wx.ID_YES, wx.ID_NO):
        raise AssertionError("dlg.ShowModal failed with %d" % status)
    ans = status == wx.ID_YES
    review_ans(dlg, ans, review)
    dlg.Destroy()
    return ans

def ask_integer(question, match=None, review=None):
    return get_answer(question,
                      qa_helpers.match_prompt(match, int, "Enter Integer %s",
                                              'Enter Integer'),
                      conv_fn=qa_helpers.to_int,
                      test=match,
                      review=review)

def ask_float(question, match=None, review=None):
    return get_answer(question,
                      qa_helpers.match_prompt(match, float, "Enter Number %s",
                                              'Enter Number'),
                      conv_fn=qa_helpers.to_float,
                      test=match,
                      review=review)

def ask_number(question, match=None, review=None):
    return get_answer(question,
                      qa_helpers.match_prompt(match, int, "Enter Number %s",
                                              'Enter Number'),
                      conv_fn=qa_helpers.to_number,
                      test=match,
                      review=review)

def ask_string(question, match=None, review=None):
    return get_answer(question,
                      qa_helpers.match_prompt(match, str, "Enter %s",
                                              'User Question'),
                      test=match,
                      review=review)

def ask_select_1(question, alternatives, review=None):
    dlg = wx.SingleChoiceDialog(None, question, 'User Question',
                                [msg for tag, msg in alternatives],
                                wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER |
                                wx.OK)
    status = dlg.ShowModal()
    if status != wx.ID_OK:
        raise AssertionError("dlg.ShowModal failed with %d" % status)
    selection = dlg.GetSelection()
    #print 'Got selection: %s' % str(selection)
    ans = alternatives[selection][0]
    review_ans(dlg, ans, review)
    dlg.Destroy()
    return ans

def ask_select_n(question, alternatives, review=None):
    #dlg = wx.lib.dialogs.MultiChoiceDialog(None, question, u'User Question',
    #                                       [msg for tag, msg in alternatives])
    dlg = wx.MultiChoiceDialog(None, question, 'User Question',
                               [msg for tag, msg in alternatives],
                               wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER |
                               wx.OK)
    status = dlg.ShowModal()
    if status != wx.ID_OK:
        raise AssertionError("dlg.ShowModal failed with %d" % status)
    #selections = dlg.GetValue()
    selections = dlg.GetSelections()
    #print 'Got selections: %s' % str(selections)
    ans = tuple(alternatives[i][0] for i in selections)
    review_ans(dlg, ans, review)
    dlg.Destroy()
    return ans




# $Id: knowledge_engine.py 2922d107410e 2010-04-26 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import sys
import types
import os, os.path
import imp
import re
import contextlib

if sys.version_info[0] < 3:
    import itertools
    class chain(object):
        old_chain = itertools.chain
        def __new__(cls, *args):
            return cls.old_chain(*args)

        @staticmethod
        def from_iterable(i):
            for iterable in i:
                for x in iterable: yield x
    itertools.chain = chain

import pyke
from pyke import contexts

debug = False

Sys_path = tuple(os.getcwd() if p == ''
                             else os.path.normpath(os.path.abspath(p))
                 for p in sys.path)

class CanNotProve(Exception):
    pass

class engine(object):
    _Variables = tuple(contexts.variable('ans_%d' % i) for i in range(100))

    def __init__(self, *search_paths, **kws):
        r'''All search_paths are relative to reference_path.

        Each search_path may be:

            path        -- a path relative to reference_path to search for
                           source files, placing the compiled knowledge bases
                           in '.compiled_krb'.
            module      -- the module's __file__ is taken as the path.
            (None|path|module, target_package)
                        -- use target_package rather than '.compiled_krb'.
                           This is a package name in Python dotted name
                           notation relative to path.  Use None to use the
                           compiled knowledge bases in the target_package
                           without scanning for source files.

        kws can be: load_fc, load_bc, load_fb and load_qb.  They are all
        boolean valued and default to True.
        '''

        # import this stuff here to avoid import cycles...
        global condensedPrint, pattern, fact_base, goal, rule_base, special, \
               target_pkg
        from pyke import (condensedPrint, pattern, fact_base, goal, rule_base,
                          special, target_pkg)

        for keyword in kws.keys():
            if keyword not in ('load_fc', 'load_bc', 'load_fb', 'load_qb'):
                raise TypeError("engine.__init__() got an unexpected keyword "
                                "argument %r" %
                                  keyword)
        self.knowledge_bases = {}
        self.rule_bases = {}
        special.create_for(self)

        if len(search_paths) == 1 and isinstance(search_paths[0], tuple) and \
           search_paths[0][0] == '*direct*' and \
           isinstance(search_paths[0][1], types.ModuleType):
            # secret hook for the compiler to initialize itself (so the
            # compiled python module can be in an egg).
            search_paths[0][1].populate(self)
        else:
            target_pkgs = {}  # {target_package_name: target_pkg}
            for path in search_paths:
                self._create_target_pkg(path, target_pkgs)
            for target_package in target_pkgs.values():
                if debug:
                    print("target_package:", target_package, file=sys.stderr)
                target_package.compile(self)
                target_package.write()
                target_package.load(self, **kws)
        for kb in self.knowledge_bases.values(): kb.init2()
        for rb in self.rule_bases.values(): rb.init2()

    def _create_target_pkg(self, path, target_pkgs):
        # Does target_pkg.add_source_package.

        if debug: print("engine._create_target_pkg:", path, file=sys.stderr)

        # First, figure out source_package_name, source_package_dir
        #               and target_package_name:
        target_package_name = '.compiled_krb'   # default
        if isinstance(path, (tuple, list)):
            path, target_package_name = path
        if isinstance(path, types.ModuleType):
            path = path.__file__
        if not isinstance(path, (str, type(None))):
            raise ValueError("illegal path argument: string expected, got " + \
                               str(type(path)))

        if debug:
            print("_create_target_pkg path:", \
                                 repr(path), file=sys.stderr)
            print("_create_target_pkg target_package_name:", \
                                 repr(target_package_name), file=sys.stderr)

        # Handle the case where there are no source files (for a distributed
        # app that wants to hide its knowledge bases):
        if path is None:
            assert target_package_name[0] != '.', \
                   "engine: relative target, %s, illegal " \
                   "with no source package" % \
                       target_package_name
            if target_package_name not in target_pkgs:
                # This import must succeed!
                tp = _get_target_pkg(target_package_name + 
                                       '.compiled_pyke_files')
                if tp is None:
                    raise AssertionError("%s: compiled with different version "
                                             "of Pyke" %
                                           target_package_name)
                tp.reset(check_sources=False)
                target_pkgs[target_package_name] = tp
            return

        path = os.path.normpath(os.path.abspath(path))

        path_to_package, source_package_name, remainder_path, zip_file_flag = \
          _pythonify_path(path)

        if debug:
            print("_create_target_pkg path to " \
                                   "_pythonify_path:", \
                                 repr(path), file=sys.stderr)
            print("    path_to_package:", repr(path_to_package), file=sys.stderr)
            print("    source_package_name:", \
                                 repr(source_package_name), file=sys.stderr)
            print("    remainder_path:", repr(remainder_path), file=sys.stderr)
            print("    zip_file_flag:", zip_file_flag, file=sys.stderr)

        target_filename = None

        # Convert relative target_package_name (if specified) to absolute form:
        if target_package_name[0] == '.':
            num_dots = \
                len(target_package_name) - len(target_package_name.lstrip('.'))
            if debug:
                print("_create_target_pkg num_dots:", num_dots, file=sys.stderr)
            if num_dots == 1:
                base_package = source_package_name
            else:
                base_package = \
                    '.'.join(source_package_name.split('.')[:-(num_dots - 1)])
            if base_package:
                target_package_name = \
                    base_package + '.' + target_package_name[num_dots:]
            else:
                target_package_name = target_package_name[num_dots:]

            target_filename = \
              os.path.join(path_to_package,
                           os.path.join(*target_package_name.split('.')),
                           'compiled_pyke_files.py')

            if debug:
                print("_create_target_pkg " \
                                       "absolute target_package_name:", \
                                     target_package_name, file=sys.stderr)

        if target_package_name in target_pkgs:
            tp = target_pkgs[target_package_name]
        else:
            target_name = target_package_name + '.compiled_pyke_files'
            if debug:
                print("_create_target_pkg target_name:", \
                                     target_name, file=sys.stderr)
            tp = None
            try:
                # See if compiled_pyke_files already exists.
                tp = _get_target_pkg(target_name)
            except ImportError:
                pass
            if tp is None:
                if debug:
                    print("_create_target_pkg: no target module", file=sys.stderr)
                tp = target_pkg.target_pkg(target_name, target_filename)
            tp.reset()
            target_pkgs[target_package_name] = tp

        source_package_dir = \
          os.path.join(path_to_package,
                       os.path.join(*source_package_name.split('.')))
        if not os.path.isdir(source_package_dir):
            source_package_dir = os.path.dirname(source_package_dir)
            remainder_path = os.path.dirname(remainder_path)
        tp.add_source_package(source_package_name, remainder_path,
                              source_package_dir)

    def get_ask_module(self):
        if not hasattr(self, 'ask_module'):
            from pyke import ask_tty
            self.ask_module = ask_tty
        return self.ask_module

    def reset(self):
        r'''Erases all case-specific facts and deactivates all rule bases.
        '''
        for rb in self.rule_bases.values(): rb.reset()
        for kb in self.knowledge_bases.values(): kb.reset()

    def get_kb(self, kb_name, _new_class = None):
        ans = self.knowledge_bases.get(kb_name)
        if ans is None:
            if _new_class: ans = _new_class(self, kb_name)
            else: raise KeyError("knowledge_base %s not found" % kb_name)
        return ans

    def get_rb(self, rb_name):
        ans = self.rule_bases.get(rb_name)
        if ans is None: raise KeyError("rule_base %s not found" % rb_name)
        return ans

    def get_create(self, rb_name, parent = None, exclude_list = ()):
        ans = self.rule_bases.get(rb_name)
        if ans is None:
            ans = rule_base.rule_base(self, rb_name, parent, exclude_list)
        elif ans.parent != parent or ans.exclude_set != frozenset(exclude_list):
            raise AssertionError("duplicate rule_base: %s" % rb_name)
        return ans

    def get_ke(self, kb_name, entity_name):
        return self.get_kb(kb_name).get_entity_list(entity_name)

    def add_universal_fact(self, kb_name, fact_name, args):
        r'''Universal facts are not deleted by engine.reset.
        '''
        if isinstance(args, str):
            raise TypeError("engine.add_universal_fact: "
                            "illegal args type, %s" % type(args))
        args = tuple(args)
        return self.get_kb(kb_name, fact_base.fact_base) \
                   .add_universal_fact(fact_name, args)

    def add_case_specific_fact(self, kb_name, fact_name, args):
        r'''Case specific facts are deleted by engine.reset.
        '''
        if isinstance(args, str):
            raise TypeError("engine.add_case_specific_fact: "
                            "illegal args type, %s" % type(args))
        args = tuple(args)
        return self.get_kb(kb_name, fact_base.fact_base) \
                   .add_case_specific_fact(fact_name, args)

    def assert_(self, kb_name, entity_name, args):
        if isinstance(args, str):
            raise TypeError("engine.assert_: "
                            "illegal args type, %s" % type(args))
        args = tuple(args)
        return self.get_kb(kb_name, fact_base.fact_base) \
                   .assert_(entity_name, args)

    def activate(self, *rb_names):
        r'''Activate rule bases.

        This runs all forward-chaining rules in the activated rule bases, so
        add your facts before doing this!
        '''
        for rb_name in rb_names: self.get_rb(rb_name).activate()

    def lookup(self, kb_name, entity_name, pat_context, patterns):
        return self.get_kb(kb_name).lookup(pat_context, pat_context,
                                           entity_name, patterns)

    def prove_goal(self, goal_str, **args):
        r'''Proves goal_str with logic variables set to args.

        This returns a context manager that you use in a with statement:

            Ugly setup to use the family_relations example.  You can ignore
            this... :-(

            >>> source_dir = os.path.dirname(os.path.dirname(__file__))
            >>> family_relations_dir = \
            ...   os.path.join(source_dir, 'examples/family_relations')
            >>> sys.path.insert(0, family_relations_dir)
            >>> from pyke import knowledge_engine
            >>> my_engine = knowledge_engine.engine(family_relations_dir)

            >>> my_engine.activate('bc_example')

            OK, here's the example!

            >>> with my_engine.prove_goal(
            ...        'family.how_related($person1, $person2, $how_related)',
            ...        person1='bruce') as it:
            ...     for vars, plan in it:
            ...         print("bruce is related to", vars['person2'], "as", \
            ...               vars['how_related'])

        vars is a dictionary of all of the logic variables in the goal
        (without the '$') and their values.  The plan is a callable python
        function.

        If you only want the first answer, see engine.prove_1_goal.
        '''
        return goal.compile(goal_str).prove(self, **args)

    def prove_1_goal(self, goal_str, **args):
        r'''Proves goal_str with logic variables set to args.

        Returns the vars and plan for the first solution found.  Raises
        knowledge_engine.CanNotProve if no solutions are found.

            Ugly setup to use the family_relations example.  You can ignore
            this... :-(

            >>> source_dir = os.path.dirname(os.path.dirname(__file__))
            >>> family_relations_dir = \
            ...   os.path.join(source_dir, 'examples/family_relations')
            >>> sys.path.insert(0, family_relations_dir)
            >>> from pyke import knowledge_engine
            >>> my_engine = knowledge_engine.engine(family_relations_dir)

            >>> my_engine.activate('bc_example')

            OK, here's the example!

            >>> vars, plan = \
            ...   my_engine.prove_1_goal(
            ...     'bc_example.how_related($person1, $person2, $how_related)',
            ...     person1='bruce',
            ...     person2='m_thomas')
            >>> print("bruce is related to m_thomas as", vars['how_related'])
            bruce is related to m_thomas as ('father', 'son')

        If you want more than one answer, see engine.prove_goal.
        '''
        return goal.compile(goal_str).prove_1(self, **args)

    def prove(self, kb_name, entity_name, pat_context, patterns):
        r'''Deprecated.  Use engine.prove_goal.
        '''
        return self.get_kb(kb_name).prove(pat_context, pat_context,
                                          entity_name, patterns)

    def prove_n(self, kb_name, entity_name, fixed_args = (), num_returns = 0):
        '''Returns a context manager for a generator of:
               a tuple of len == num_returns, and a plan (or None).

        Deprecated.  Use engine.prove_goal.
        '''
        if isinstance(fixed_args, str):
            raise TypeError("engine.prove_n: fixed_args must not be a string, "
                            "did you forget a , (%(arg)s) => (%(arg)s,)?" %
                            {'arg': repr(fixed_args)})
        def gen():
            context = contexts.simple_context()
            vars = self._Variables[:num_returns]
            try:
                with self.prove(kb_name, entity_name, context,
                                tuple(pattern.pattern_literal(arg)
                                      for arg in fixed_args) + vars) \
                  as it:
                    for plan in it:
                        final = {}
                        ans = tuple(context.lookup_data(var.name, final = final)
                                    for var in vars)
                        if plan: plan = plan.create_plan(final)
                        yield ans, plan
            finally:
                context.done()
        return contextlib.closing(gen())

    def prove_1(self, kb_name, entity_name, fixed_args = (), num_returns = 0):
        r'''Returns a tuple of len == num_returns, and a plan (or None).

        Deprecated.  Use engine.prove_1_goal.
        '''
        try:
            # All we need is the first one!
            with self.prove_n(kb_name, entity_name, fixed_args, num_returns) \
              as it:
                return next(iter(it))
        except StopIteration:
            raise CanNotProve("Can not prove %s.%s%s" %
                               (kb_name, entity_name,
                                 condensedPrint.cprint(
                                   fixed_args + self._Variables[:num_returns])))

    def print_stats(self, f = sys.stdout):
        for kb \
         in sorted(iter(self.knowledge_bases.values()), key=lambda kb: kb.name):
            kb.print_stats(f)

    def trace(self, rb_name, rule_name):
        self.get_rb(rb_name).trace(rule_name)

    def untrace(self, rb_name, rule_name):
        self.get_rb(rb_name).untrace(rule_name)

Compiled_suffix = None

def _get_target_pkg(target_name):
    global Compiled_suffix
    if debug: print("_get_target_pkg", target_name, file=sys.stderr)
    module = target_pkg.import_(target_name)
    path = module.__file__
    if debug: print("_get_target_pkg __file__ is", path, file=sys.stderr)
    do_reload = False
    if path.endswith('.py'):
        if Compiled_suffix is None:
            # We don't know whether this compiles to .pyc or .pyo yet, so do a
            # reload just to be sure...
            do_reload = True
        else:
            source_path = path
            path = path[:-3] + Compiled_suffix
    else:
        assert path.endswith(('.pyc', '.pyo')), \
               'unknown file extension: %r' % (path,)
        Compiled_suffix = path[-4:]
        source_path = path[:-1]
    if not do_reload:
        if debug:
            print("source path is", source_path, file=sys.stderr)
            if os.path.exists(source_path):
                print("source path exists", file=sys.stderr)
                print("source path mtime", \
                                     os.path.getmtime(source_path), file=sys.stderr)
            else:
                print("source path does not exist", file=sys.stderr)
            print("compiled path is", path, file=sys.stderr)
            if os.path.exists(path):
                print("compiled path exists", file=sys.stderr)
                print("compiled path mtime", \
                                     os.path.getmtime(path), file=sys.stderr)
            else:
                print("compiled path does not exist", file=sys.stderr)
        if not os.path.exists(path) or \
              os.path.exists(source_path) and \
              os.path.getmtime(source_path) > os.path.getmtime(path):
            do_reload = True
    if do_reload:
        if debug:
            print("_get_target_pkg doing reload for", target_name, file=sys.stderr)
        module = imp.reload(module)
        suffix = module.__file__[-4:]
        if suffix in ('.pyc', '.pyo'):
            Compiled_suffix = suffix
    if getattr(module, 'target_pkg_version', None) != pyke.target_pkg_version:
        if debug:
            print("_get_target_pkg doing invalid version for", \
                                 target_name, file=sys.stderr)
        return None
    return getattr(module, 'get_target_pkg')()

def _pythonify_path(path):
    r'''Returns path_to_package, package_name, remainder_path, zip_file_flag.

    If zip_file_flag is set, remainder_path is ''.
    '''
    path = os.path.normpath(os.path.abspath(path))
    if path.endswith(('.py', '.pyw', '.pyc', '.pyo')):
        path = os.path.dirname(path)
    package_name = ''
    remainder_path = ''
    remainder_package_name = ''
    ans = '', '', path, False
    while path:
        if in_sys_path(path):
            if len(remainder_path) < len(ans[2]) or \
               len(remainder_path) == len(ans[2]) and \
                 len(package_name) > len(ans[1]):
                if os.path.isdir(path):
                    ans = path, package_name, remainder_path, False
                else:
                    ans = path, remainder_package_name, '', True
        parent_path, dir = os.path.split(path)
        if parent_path == '' or parent_path == path:
            break
        if _is_package_dir(path):
            if package_name:
                package_name = dir + '.' + package_name
            else:
                package_name = dir
        else:
            package_path = os.path.join(*package_name.split('.'))
            package_name = ''
            if remainder_path:
                remainder_path = os.path.join(dir, package_path, remainder_path)
            else:
                remainder_path = os.path.join(dir, package_path)
        if remainder_package_name:
            remainder_package_name = dir + '.' + remainder_package_name
        else:
            remainder_package_name = dir
        path = parent_path
    return ans

def _is_package_dir(path):
    if not os.path.isdir(path): return False
    return os.path.exists(os.path.join(path, '__init__.py')) or \
           os.path.exists(os.path.join(path, '__init__.pyw')) or \
           os.path.exists(os.path.join(path, '__init__.pyc')) or \
           os.path.exists(os.path.join(path, '__init__.pyo'))

def in_sys_path(path):
    r'''Assumes path is a normalized abspath.
    '''
    return path in Sys_path





# $Id: __init__.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import os, os.path
import sys

import pyke
from pyke import knowledge_engine

from pyke.krb_compiler import compiler_bc
from pyke.krb_compiler import krbparser

#from pyke import contexts
#contexts.debug = ('patterns_out1', 'patterns_out',)

Ast_names = frozenset((
    'file',
    'parent',
    'fc_rule',
    'fc_predicate',
    'assert',
    'python_assertion',
    'python_eq',
    'python_in',
    'python_check',
    'counting',
    'bc_rule',
    'goal',
    'bc_predicate',
    'symbol',
    'pattern_var',
    'as',
    'plan_spec',
    'pattern_data',
    'pattern_tuple',
  # 'anonymous_var',
))

def dump(ast, f = sys.stderr, need_nl = False, indent = 0):
    if not isinstance(ast, tuple) or len(ast) == 0:
        f.write(repr(ast))
        return False
    if ast[0] in Ast_names:
        indent += 2
        if need_nl:
            f.write("\n")
            f.write(' ' * indent) 
        f.write('(%s' % ast[0])
        for arg in ast[1:]:
            f.write(', ')
            dump(arg, f, True, indent)
        f.write(')')
        return True
    f.write('(')
    did_nl = dump(ast[0], f, False, indent)
    for arg in ast[1:]:
        f.write(', ')
        did_nl |= dump(arg, f, did_nl, indent)
    f.write(')')
    return did_nl

def to_relative(from_path, to_path):
    '''Calculates the relative path to get from from_path to to_path.

        >>> to_relative('/a/b/c', '/a/b/d/e')
        '../d/e'
        >>> to_relative('/a/b/c', '/b/d/e')
        '/b/d/e'
        >>> to_relative('/a/b/c', '/a/b/c/e')
        'e'
        >>> to_relative('/a/b/c', '/a/b2/d/e')
        '../../b2/d/e'
    '''
    from_path = os.path.abspath(from_path)
    to_path = os.path.abspath(to_path)
    prefix = ''
    while os.path.join(from_path, to_path[len(from_path) + 1:]) != to_path:
        new_from_path = os.path.dirname(from_path)
        if new_from_path == from_path: return to_path
        from_path = new_from_path
        prefix = os.path.join(prefix, '..')
    return os.path.join(prefix, to_path[len(from_path) + 1:])

def compile_krb(rb_name, generated_root_pkg, generated_root_dir, filename):
    engine = knowledge_engine.engine(('*direct*', compiler_bc))
    try:
        fc_name = rb_name + '_fc.py'
        bc_name = rb_name + '_bc.py'
        plan_name = rb_name + '_plans.py'
        fc_path = os.path.join(generated_root_dir, fc_name)
        bc_path = os.path.join(generated_root_dir, bc_name)
        plan_path = os.path.join(generated_root_dir, plan_name)
        ast = krbparser.parse(krbparser, filename)
        #sys.stderr.write("got ast\n")
        # dump(ast)
        # sys.stderr.write('\n\n')
        engine.reset()
        engine.activate('compiler')
        (fc_lines, bc_lines, plan_lines), plan = \
            engine.prove_1('compiler', 'compile',
                           (generated_root_pkg, rb_name, ast), 3)
        krb_filename = to_relative(generated_root_dir, filename)
        ans = []
        if fc_lines:
            sys.stderr.write("writing [%s]/%s\n" %
                               (generated_root_pkg, os.path.basename(fc_path)))
            write_file(fc_lines +
                       ("",
                        "Krb_filename = %r" % krb_filename,),
                       fc_path)
            ans.append(fc_name)
        elif os.path.lexists(fc_path): os.remove(fc_path)
        if bc_lines:
            sys.stderr.write("writing [%s]/%s\n" %
                               (generated_root_pkg, os.path.basename(bc_path)))
            write_file(bc_lines +
                       ("",
                        "Krb_filename = %r" % krb_filename,),
                       bc_path)
            ans.append(bc_name)
        elif os.path.lexists(bc_path): os.remove(bc_path)
        if plan_lines:
            sys.stderr.write("writing [%s]/%s\n" %
                               (generated_root_pkg,
                                os.path.basename(plan_path)))
            #sys.stderr.write("plan_lines:\n")
            #for line in plan_lines:
            #    sys.stderr.write("  " + repr(line) + "\n")
            write_file(plan_lines +
                       ("",
                        "Krb_filename = %r" % krb_filename,),
                       plan_path)
            ans.insert(len(ans) - 1, plan_name)  # want this loaded before _bc
        elif os.path.lexists(plan_path): os.remove(plan_path)
        #sys.stderr.write("done!\n")
        return ans
    except:
        if os.path.lexists(fc_path): os.remove(fc_path)
        if os.path.lexists(bc_path): os.remove(bc_path)
        if os.path.lexists(plan_path): os.remove(plan_path)
        raise

def compile_goal(goal_str):
    return krbparser.parse_goal(krbparser, goal_str)

def compile_kfb(filename):
    global kfbparser
    try:
        kfbparser
    except NameError:
        from pyke.krb_compiler import kfbparser
    return kfbparser.parse(kfbparser, filename)

def compile_kqb(filename):
    global kqb_parser
    try:
        kqb_parser
    except NameError:
        from pyke.krb_compiler import kqb_parser
    return kqb_parser.parse_kqb(filename)

def write_file(lines, filename):
    with open(filename, 'w') as f:
        indents = [0]
        lineno_map = []
        write_file2(lines, f, indents, lineno_map, 0)
        if lineno_map:
            f.write("Krb_lineno_map = (\n")
            for map_entry in lineno_map:
                f.write("    %s,\n" % str(map_entry))
            f.write(")\n")

def write_file2(lines, f, indents, lineno_map, lineno, starting_lineno = None):
    for line in lines:
        if line == 'POPINDENT':
            assert len(indents) > 1
            del indents[-1]
        elif isinstance(line, tuple):
            if len(line) == 2 and line[0] == 'INDENT':
                indents.append(indents[-1] + line[1])
            elif len(line) == 2 and line[0] == 'STARTING_LINENO':
                assert starting_lineno is None, \
                       "missing ENDING_LINENO for STARTING_LINENO %d" % \
                           starting_lineno[1]
                starting_lineno = line[1], lineno + 1
            elif len(line) == 2 and line[0] == 'ENDING_LINENO':
                assert starting_lineno is not None, \
                       "missing STARTING_LINENO for ENDING_LINENO %d" % \
                           line[1]
                lineno_map.append(((starting_lineno[1], lineno),
                                   (starting_lineno[0], line[1])))
                starting_lineno = None
            else:
                lineno, starting_lineno = \
                    write_file2(line, f, indents, lineno_map, lineno,
                                starting_lineno)
        else:
            f.write(' ' * indents[-1] + line + '\n')
            lineno += 1
    return lineno, starting_lineno

